#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include <inttypes.h>

typedef uint64_t count_t ;
#define PRC PRIu64

static int verbose = 0 ;


/* Barrier */


#include "sense_barrier.c"


typedef struct {
  int size ;
  barrier_t *b ;
  volatile int *x,*y;
  volatile int *r1,*r2,*r3,*r4 ;
} common_t ;

inline static void free_tab(volatile int *p) { free((void *)p) ; }

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

inline static void lwsync(void) {
  //  asm __volatile__ ("lwsync" ::: "memory") ;
}

void *T1(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_my_barrier(q->b,&sense) ;
    q->x[k] = 1 ;
  }
  return NULL ;
}

void *T2(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_my_barrier(q->b,&sense) ;
    int r1 = q->x[k] ;
    lwsync() ;
    int r2 = q->y[k] ;

    q->r1[k] = r1 ;
    q->r2[k] = r2 ;
  }
  return NULL ;
}

void *T3(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_my_barrier(q->b,&sense) ;
    q->y[k] = 1 ;
  }
  return NULL ;
}

void *T4(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_my_barrier(q->b,&sense) ;
    int r3 = q->y[k] ;
    lwsync() ;
    int r4 = q->x[k] ;

    q->r3[k] = r3 ;
    q->r4[k] = r4 ;
  }
  return NULL ;
}

typedef count_t c_t[2][2][2][2] ;

void init_count(c_t count) {
  for (int k1 = 0 ; k1 < 2 ; k1++)
    for (int k2 = 0 ; k2 < 2 ; k2++)
      for (int k3 = 0 ; k3 < 2 ; k3++)
        for (int k4 = 0 ; k4 < 2 ; k4++)
          count[k1][k2][k3][k4] = 0 ;
}

void dump_count(c_t count) {
  for (int k1 = 0 ; k1 < 2 ; k1++)
    for (int k2 = 0 ; k2 < 2 ; k2++)
      for (int k3 = 0 ; k3 < 2 ; k3++)
	for (int k4 = 0 ; k4 < 2 ; k4++) {
	  count_t c = count[k1][k2][k3][k4] ;
	  printf("%-5" PRC "> r1=%i, r2=%i, r3=%i, r4=%i\n",c,k1,k2,k3,k4) ;
	}
}

typedef struct {
  int size,runmax ;
  c_t count ;
  pthread_mutex_t *mutex ;
  barrier_t * b ;
} cr_t ;

typedef struct {
  int id ;
  cr_t *common ;
} run_t ;


void *run(void *_p) {
  run_t *p = _p ; cr_t *q = p->common ;
  int size = q->size ;
  int runmax = q->runmax ;

  c_t count ; init_count(count) ;
  common_t c ;
  c.x = malloc_check(size * sizeof(c.x[0])) ;
  c.y = malloc_check(size * sizeof(c.y[0])) ;
  c.r1 = malloc_check(size * sizeof(c.r1[0])) ;
  c.r2 = malloc_check(size * sizeof(c.r2[0])) ;
  c.r3 = malloc_check(size * sizeof(c.r3[0])) ;
  c.r4 = malloc_check(size * sizeof(c.r4[0])) ;
  c.size = size ; c.b = alloc_my_barrier(4) ;

  for (int nrun = 1 ; nrun <= runmax ; nrun++) {
    for (int k = 0 ; k < size ; k++) {
      c.x[k] = c.y[k] = 0 ;
      c.r1[k] = c.r2[k] = c.r3[k] = c.r4[k] = -1 ;
    }
    pthread_t th1,th2,th3,th4 ;
    ctx_t a1,a2,a3,a4 ;
    
    a1.id = 1 ; a1.common = &c ;
    create_thread(&th1,T1,&a1) ;
    a2.id = 2 ; a2.common = &c ;
    create_thread(&th2,T2,&a2) ;
    a3.id = 3 ; a3.common = &c ;
    create_thread(&th3,T3,&a3) ;
    a4.id = 4 ; a4.common = &c ;
    create_thread(&th4,T4,&a4) ;

    (void)join_thread(&th1) ;
    (void)join_thread(&th2) ;
    (void)join_thread(&th3) ;
    (void)join_thread(&th4) ;
    for (int k = 0 ; k < size ; k++)
      count[c.r1[k]][c.r2[k]][c.r3[k]][c.r4[k]]++ ;
  }
  free_tab(c.x) ; free_tab(c.y) ;
  free_tab(c.r1) ; free_tab(c.r2) ; free_tab(c.r3) ; free_tab(c.r4) ;
  free_my_barrier(c.b) ;

  lock_mutex(q->mutex) ;
  for (int k1 = 0 ; k1 < 2 ; k1++)
    for (int k2 = 0 ; k2 < 2 ; k2++)
      for (int k3 = 0 ; k3 < 2 ; k3++)
	for (int k4 = 0 ; k4 < 2 ; k4++) {
	  q->count[k1][k2][k3][k4] += count[k1][k2][k3][k4] ;
	}
  unlock_mutex(q->mutex) ;
  return NULL ;
}



void zyva(int size, int ninst) {
  int nrun = 0 ;
  cr_t c ;
  c.size = size ;
  c.runmax = 10 ;
  c.mutex = alloc_mutex() ;
  c.b = alloc_my_barrier(ninst * 4) ;
  init_count(c.count) ;

  for ( ; ; ) {
    nrun++ ;
    pthread_t th[ninst] ; run_t a[ninst] ;
    for (int k = 0 ; k < ninst ; k++) {
      run_t *p = &a[k] ; p->id = k ; p->common = &c ;
      create_thread(&th[k],run,p) ;
    }
    for (int k = 0 ; k < ninst ; k++) (void)join_thread(&th[k]) ;
    printf("** %i\n",nrun) ;
    dump_count(c.count) ;
  }
  free_mutex(c.mutex) ;
  free_my_barrier(c.b) ;
}

void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [size] [ninst]\n",pgm) ;
}


int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 100000 ;
  int ninst = 1 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc > 1) {
    size = atoi(argv[0]) ;
    ninst = atoi(argv[1]) ;
  } if (argc == 1) {    
    size = atoi(argv[0]) ;
  }
  zyva(size,ninst) ;
  return 0 ;
}
