#include <stdlib.h>
#include <stdio.h>
#include "basic.h"

#define CAS(x,y,z) __sync_bool_compare_and_swap(x,y,z)
/*
Treiber's stack

    void push (int v, stack &S)
        node *n = new node
        n->val = v
        while (1)
            node *top = S.top
            n->next = top
            if cas(&S.top, top, n)
                return

    bool pop (stack &S, int &rtn)
        while (1)
            node *top = S.top
            if top == nil
                return FALSE
            if cas(&S.top, top, top->next)
                rtn = top->val
                free top
                return TRUE

*/

typedef struct {
  volatile int sz,sp ;
  int *t ;
} stack_t ;

stack_t *alloc_stack(int sz) {
  stack_t *r = malloc_check(sizeof(*r)) ;
  r->sz = sz ;
  r->sp = 0 ;
  r->t = calloc_check(sz,sizeof(r->t[0])) ;
  return r ;  
}

void free_stack(stack_t *p) {
  free(p->t)  ;
  free(p) ;
}

int push(stack_t *p,int v) {
  for ( ; ; ) {
    int sp = p->sp ;
    if (sp >= p->sz) return 1 ;
    if (CAS(&p->sp,sp,sp+1)) {
      p->t[sp] = v ;
      return 0 ;
    }
  }  
}

int pop(stack_t *p,int *ret) {
  for ( ; ; ) {
    int sp = p->sp ;
    if (sp == 0) return 1 ;
    int v = p->t[sp-1] ;
    if (CAS(&p->sp,sp,sp-1)) {
      *ret = v ;
      return 0 ;
    }
  }
}

static int n = 1000 ;
static stack_t *stack ;

void *pusher(void *p) {
  for (int k = 0 ; k < n ; k++) {
    while (push(stack,k+1)) ;
  }
  return NULL ;
}


void *popper(void *p) {
  int sum = 0 ;
  for (int k = 0 ; k < n ; k++) {
    int v ;
    while (pop(stack,&v)) ;
    sum += v ;
  }
  return alloc_boxed_int(sum) ;
}

int seq(int n) {
  int sum = 0 ;
  for (int k = 0 ; k < n ; k++) sum += k+1 ;
  return sum ;
}
void run(void) {
  stack = alloc_stack(10) ;

  pthread_t th1,th2 ;

  create_thread(&th1,pusher,NULL) ;
  create_thread(&th2,popper,NULL) ;
  (void)join_thread(&th1) ;
  boxed_int_t *r = join_thread(&th2) ;

  int ok = seq(n) ;
  printf("%i %i\n",r->v, ok) ;

  free_stack(stack) ;
}

int main(int argc, char **argv) {
  if (argc > 1) n = atoi(argv[1]) ;
  run() ;
  return 0 ;
}
